//////////////////////////////////////////////////////////////////////
// Simple ARM7 stub (sends RTC, TSC, and X/Y data to the ARM 9)
// -- joat
// -- modified by Darkain and others
//////////////////////////////////////////////////////////////////////


#include <stdlib.h>
#include <stdio.h>
 
#include "nds_compati.h"
//#include <NDS.h>
#include <nds/ipc.h>

#include <NDS/ARM7/touch.h>
#include <NDS/ARM7/clock.h>
#include <nds/registers_alt.h>
#include "../../ipc3.h"
#include "memtoolARM7.h"

#include "a7sleep.h"



#include <nds/jtypes.h>
#include <nds/system.h>
#include <nds/arm7/touch.h>


//---------------------------------------------------------------------------------
uint16 touchRead(uint32 command) {
//---------------------------------------------------------------------------------
	uint16 result;
	SerialWaitBusy();

	// Write the command and wait for it to complete
	REG_SPICNT = SPI_ENABLE | SPI_BAUD_2MHz | SPI_DEVICE_TOUCH | SPI_CONTINUOUS; //0x0A01;
	REG_SPIDATA = command;
	SerialWaitBusy();

	// Write the second command and clock in part of the data
	REG_SPIDATA = 0;
	SerialWaitBusy();
	result = REG_SPIDATA;

	// Clock in the rest of the data (last transfer)
	REG_SPICNT = SPI_ENABLE | 0x201;
	REG_SPIDATA = 0;
	SerialWaitBusy();

	// Return the result
	return ((result & 0x7F) << 5) | (REG_SPIDATA >> 3);
}


//---------------------------------------------------------------------------------
uint32 touchReadTemperature(int * t1, int * t2) {
//---------------------------------------------------------------------------------
	*t1 = touchRead(TSC_MEASURE_TEMP1);
	*t2 = touchRead(TSC_MEASURE_TEMP2);
	return 8490 * (*t2 - *t1) - 273*4096;
}



static bool touchInit = false;
static s32 xscale, yscale;
static s32 xoffset, yoffset;

// reading pixel position:
//---------------------------------------------------------------------------------
touchPosition touchReadXY() {
//---------------------------------------------------------------------------------


	touchPosition touchPos;


	if ( !touchInit ) {


		xscale = ((PersonalData->calX2px - PersonalData->calX1px) << 16) / ((PersonalData->calX2 & -8) - (PersonalData->calX1 & -8));
		yscale = ((PersonalData->calY2px - PersonalData->calY1px) << 16) / ((PersonalData->calY2 & -8) - (PersonalData->calY1 & -8));

		xoffset = (PersonalData->calX1 & -8) * xscale - (PersonalData->calX1px << 16);
		yoffset = (PersonalData->calY1 & -8) * yscale - (PersonalData->calY1px << 16);

		touchInit = true;

	}


	touchPos.x = touchRead(TSC_MEASURE_X) & -8;
	touchPos.y = touchRead(TSC_MEASURE_Y) & -8;


	s16 px = ( touchPos.x * xscale - xoffset + xscale/2 ) >>16;
	s16 py = ( touchPos.y * yscale - yoffset + yscale/2 ) >>16;


	if ( px < 0) px = 0;
	if ( py < 0) py = 0;
	if ( px > (SCREEN_WIDTH -1)) px = SCREEN_WIDTH -1;
	if ( py > (SCREEN_HEIGHT -1)) px = SCREEN_HEIGHT -1;

	touchPos.px = px;
	touchPos.py = py;
	
	IPC3->touchXpx = touchPos.px;
	

	return touchPos;

}

//////////////////////////////////////////////////////////////////////


#define TOUCH_CAL_X1 (*(vs16*)0x027FFCD8)
#define TOUCH_CAL_Y1 (*(vs16*)0x027FFCDA)
#define TOUCH_CAL_X2 (*(vs16*)0x027FFCDE)
#define TOUCH_CAL_Y2 (*(vs16*)0x027FFCE0)
#define SCREEN_WIDTH    256
#define SCREEN_HEIGHT   192
s32 TOUCH_WIDTH  = TOUCH_CAL_X2 - TOUCH_CAL_X1;
s32 TOUCH_HEIGHT = TOUCH_CAL_Y2 - TOUCH_CAL_Y1;
s32 TOUCH_OFFSET_X = ( ((SCREEN_WIDTH -60) * TOUCH_CAL_X1) / TOUCH_WIDTH  ) - 28;
s32 TOUCH_OFFSET_Y = ( ((SCREEN_HEIGHT-60) * TOUCH_CAL_Y1) / TOUCH_HEIGHT ) - 28;

static u32 strpcmCursorFlag=0;
static u32 strpcmCurrentFreq=0;

static u32 strpcmSamples,strpcmChannels;
static s16 *strpcmLBuf,*strpcmRBuf;

static s16 *strpcmL0=NULL,*strpcmL1=NULL;
static s16 *strpcmR0=NULL,*strpcmR1=NULL;

#undef SOUND_FREQ
//#define SOUND_FREQ(n)	(0x10000 - (0x1000000 / (n)))
//#define SOUND_FREQ(n)	(0x10000 - (16780000 / (n)))
#define SOUND_FREQ(n) (0x10000 - (16756000 / (n)))


void strpcmPlay() {
  strpcmCursorFlag=0;
  
  strpcmCurrentFreq=SOUND_FREQ(IPC3->strpcmFreq);
  
  strpcmSamples=IPC3->strpcmSamples;
  strpcmChannels=IPC3->strpcmChannels;
  
  strpcmLBuf=IPC3->strpcmLBuf;
  strpcmRBuf=IPC3->strpcmRBuf;
  
  strpcmL0=(s16*)malloc(strpcmSamples*2);
  strpcmL1=(s16*)malloc(strpcmSamples*2);
  strpcmR0=(s16*)malloc(strpcmSamples*2);
  strpcmR1=(s16*)malloc(strpcmSamples*2);
  
  MemSet16DMA3(0,strpcmL0,strpcmSamples*2);
  MemSet16DMA3(0,strpcmL1,strpcmSamples*2);
  MemSet16DMA3(0,strpcmR0,strpcmSamples*2);
  MemSet16DMA3(0,strpcmR1,strpcmSamples*2);
  
//  powerON(POWER_SOUND);
//  SOUND_CR = SOUND_ENABLE | SOUND_VOL(0x7F);
  SCHANNEL_CR(0) = 0;
  SCHANNEL_CR(1) = 0;
  SCHANNEL_CR(2) = 0;
  SCHANNEL_CR(3) = 0;
  
//  TIMER0_DATA = (u16)(0x10000-((0x1000000/IPC3->strpcmFreq)*2));
TIMER0_DATA = (u16)(0x10000-((16756000/IPC3->strpcmFreq)*2));
  TIMER0_CR = TIMER_DIV_1 | TIMER_ENABLE;
  
  TIMER1_DATA = 0x10000 - strpcmSamples;
  TIMER1_CR = TIMER_CASCADE | TIMER_IRQ_REQ | TIMER_ENABLE;

}

void strpcmStop() {
//  powerOFF(POWER_SOUND);
//  SOUND_CR = 0;
  TIMER0_CR = 0;
  TIMER1_CR = 0;
  
  if(strpcmL0!=NULL){
    free(strpcmL0); strpcmL0=NULL;
  }
  if(strpcmL1!=NULL){
    free(strpcmL1); strpcmL1=NULL;
  }
  if(strpcmR0!=NULL){
    free(strpcmR0); strpcmR0=NULL;
  }
  if(strpcmR1!=NULL){
    free(strpcmR1); strpcmR1=NULL;
  }
}

//////////////////////////////////////////////////////////////////////
void initClock()
{
   REG_RCNT = 0x8100;


   // Reset the clock if needed
   rtcReset();

   rtcGetTimeAndDate((uint8 *)&(IPC3->time.rtc.year));

}

void syncRTC()
{
   int oldhours = IPC3->time.rtc.hours;
   
 

   rtcGetTime((uint8 *)&(IPC3->time.rtc.hours));

   if(oldhours>IPC3->time.rtc.hours) // going from 23 to 0 hours, update whole time struct
      rtcGetTimeAndDate((uint8 *)&(IPC3->time.rtc.year));
} 
vuint32 frame = 0;



void VsyncUpdate(void)
{
  uint16 but=0, x=0, y=0, xpx=0, ypx=0, z1=0, z2=0, batt=0, aux=0;
  int t1=0, t2=0;
  uint32 temp=0;

  
  // Read the X/Y buttons and the /PENIRQ line
  but = XKEYS;
  if (!(but & 0x40)) {
    // Read the touch screen
    x = touchRead(TSC_MEASURE_X);
    y = touchRead(TSC_MEASURE_Y);
    xpx = ( ((SCREEN_WIDTH -60) * x) / TOUCH_WIDTH  ) - TOUCH_OFFSET_X;
    ypx = ( ((SCREEN_HEIGHT-60) * y) / TOUCH_HEIGHT ) - TOUCH_OFFSET_Y;
    z1 = touchRead(TSC_MEASURE_Z1);
    z2 = touchRead(TSC_MEASURE_Z2);
  }
  
  batt = touchRead(TSC_MEASURE_BATTERY);
  aux  = touchRead(TSC_MEASURE_AUX);
  
  // Read the time

  
  // Read the temperature
  temp = touchReadTemperature(&t1, &t2);
  
  // Update the IPC struct
  IPC3->heartbeat++;
  IPC3->buttons   = but;
  IPC3->touchX    = x;
  IPC3->touchY    = y;
  IPC3->touchXpx  = xpx;
  IPC3->touchYpx  = ypx;
  IPC3->touchZ1   = z1;
  IPC3->touchZ2   = z2;
  IPC3->battery   = batt;
  IPC3->aux       = aux;

  IPC3->temperature = temp;
  IPC3->tdiode1 = t1;
  IPC3->tdiode2 = t2;
  
  touchReadXY();
  
  
   
}

void InterruptHandler(void) {
  u32 flags = IF;
	if ( ++frame == 60 ) { // update time struct once per second
      frame = 0;
      syncRTC();
  }
  if (flags & IRQ_TIMER1) {
    
    s16 *lbuf,*rbuf;
    
    if(strpcmCursorFlag==0){
      lbuf=strpcmL0;
      rbuf=strpcmR0;
      }else{
      lbuf=strpcmL1;
      rbuf=strpcmR1;
    }
    
    u32 channel=strpcmCursorFlag;
    
    // Left channel
    SCHANNEL_CR(channel) = 0;
    SCHANNEL_TIMER(channel) = strpcmCurrentFreq;
    SCHANNEL_SOURCE(channel) = (uint32)lbuf;
    SCHANNEL_LENGTH(channel) = (strpcmSamples * 2) >> 2;
    SCHANNEL_REPEAT_POINT(channel) = 0;
    SCHANNEL_CR(channel) = SCHANNEL_ENABLE | SOUND_ONE_SHOT | SOUND_VOL(0x7F) | SOUND_PAN(0) | SOUND_16BIT;
    
    channel+=2;
    
    // Right channel
    SCHANNEL_CR(channel) = 0;
    SCHANNEL_TIMER(channel) = strpcmCurrentFreq;
    SCHANNEL_SOURCE(channel) = (uint32)rbuf;
    SCHANNEL_LENGTH(channel) = (strpcmSamples * 2) >> 2;
    SCHANNEL_REPEAT_POINT(channel) = 0;
    SCHANNEL_CR(channel) = SCHANNEL_ENABLE | SOUND_ONE_SHOT | SOUND_VOL(0x7F) | SOUND_PAN(0x7F) | SOUND_16BIT;
    
    strpcmCursorFlag=1-strpcmCursorFlag;
    
    if(strpcmCursorFlag==0){
      lbuf=strpcmL0;
      rbuf=strpcmR0;
      }else{
      lbuf=strpcmL1;
      rbuf=strpcmR1;
    }
    
    s16 *lsrc=strpcmLBuf;
    s16 *rsrc=strpcmRBuf;
    
    u32 Samples=strpcmSamples;
    
    if((IPC_SYNC&IPC_SYNC_IRQ_REQUEST)!=0){
      MemSet16DMA3(0,lbuf,Samples*2);
      MemSet16DMA3(0,rbuf,Samples*2);
      }else{
      if(strpcmChannels==2){
        MemCopy16DMA3(lsrc,lbuf,Samples*2);
        MemCopy16DMA3(rsrc,rbuf,Samples*2);
        }else{
        MemCopy16DMA3(lsrc,lbuf,Samples*2);
      }
      
      IPC_SYNC|=IPC_SYNC_IRQ_REQUEST;
//
//	08:30 2006/01/26 ~b^[ǉ�
//
#define MAX( x, y ) ( ( x > y ) ? x : y )
#define MIN( x, y ) ( ( x < y ) ? x : y )
      if(strpcmChannels==2){
        s32 Volume=(s32)IPC3->strpcmVolume8;
        for(u32 cnt=0;cnt<Samples;cnt++){
          s32 SrcSample;
          s32 Sample;
          
          SrcSample=(s32)lbuf[cnt];
          Sample=MAX(MIN(32767,(SrcSample*Volume)/128),-32768);
          lbuf[cnt]=(s16)Sample;
          
          SrcSample=(s32)rbuf[cnt];
          Sample=MAX(MIN(32767,(SrcSample*Volume)/128),-32768);
          rbuf[cnt]=(s16)Sample;
        }
        }else{
        s32 Volume=(s32)IPC3->strpcmVolume8;
        for(u32 cnt=0;cnt<Samples;cnt++){
          s32 SrcSample;
          s32 Sample;
          
          SrcSample=(s32)lbuf[cnt];
          Sample=MAX(MIN(32767,(SrcSample*Volume)/128),-32768);
          lbuf[cnt]=rbuf[cnt]=(s16)Sample;
        }
      }

/*      
      if(strpcmChannels==2){
        u32 Volume=IPC3->strpcmVolume8;
        for(u32 cnt=0;cnt<Samples;cnt++){
          s32 Sample;
          
          Sample=((s32)lbuf[cnt]*Volume)/8;
          lbuf[cnt]=(s16)Sample;
          
          Sample=((s32)rbuf[cnt]*Volume)/8;
          rbuf[cnt]=(s16)Sample;
        }
        }else{
        u32 Volume=IPC3->strpcmVolume8;
        for(u32 cnt=0;cnt<Samples;cnt++){
          s32 Sample;
          
          Sample=((s32)lbuf[cnt]*Volume)/8;
          lbuf[cnt]=(s16)Sample;
          rbuf[cnt]=(s16)Sample;
        }
      }
*/

    }
    
    VBLANK_INTR_WAIT_FLAGS |= IRQ_TIMER1;
    IF = IRQ_TIMER1;
    return;
  }
  

  if (IF & IRQ_VBLANK) {
  	static u32 oxkeys;
  	       u32 xkeys;
    xkeys = XKEYS;
    if(oxkeys!=xkeys)
    {
    	oxkeys=xkeys;
		IPC3->ReqVsyncUpdate=1;
	    if(xkeys == 0x00FF)	//	plN[Y
	    {
			a7lcdbl(0);
	    }
	    else
	    {
			a7lcdbl(3);
			//a7lcdbl(1);
	    }
    }
    VsyncUpdate();

    VBLANK_INTR_WAIT_FLAGS |= IRQ_VBLANK;
    IF = IRQ_VBLANK;
    return;
  }
 
  // Acknowledge interrupts
  IF = IF;
  
  
}


 

//////////////////////////////////////////////////////////////////////


#define DISP_SR        (*(vuint16*)0x04000004)
int main(int argc, char ** argv) {
  // Reset the clock if needed
  initClock();

  IPC3->strpcmControl=strpcmControl_NOP;
  
  // Set up the interrupt handler
  IME = 0;
  IRQ_HANDLER = &InterruptHandler;
  IE = IRQ_TIMER1 | IRQ_VBLANK;
  DISP_SR = DISP_VBLANK_IRQ;
  IF = ~0;
  IME = 1;
  
  powerON(POWER_SOUND);
  SOUND_CR = SOUND_ENABLE | SOUND_VOL(0x7F);
  
  
  
  IPC3->heartbeat=0;
  while(IPC3->heartbeat==0)
  {
	  swiWaitForVBlank();
  }


//		WAIT_CR &= ~(0x8000);
		WAIT_CR &= ~0x8000;	//ARM7


  
  // Keep the ARM7 out of main RAM
  //a7lcdbl(1);
  while (1){

    swiWaitForVBlank();
    if(IPC3->ReqVsyncUpdate!=0){
      
      IPC3->ReqVsyncUpdate=0;
    }
    
    VsyncUpdate();;
    if(IPC3->strpcmControl!=strpcmControl_NOP){
      if(IPC3->strpcmControl==strpcmControl_Play){
        strpcmPlay();
      }
      if(IPC3->strpcmControl==strpcmControl_Stop){
        strpcmStop();
      }
      IPC3->strpcmControl=strpcmControl_NOP;
    }
  }
  return 0;
}


//////////////////////////////////////////////////////////////////////
