/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggVorbis 'TREMOR' CODEC SOURCE CODE.   *
 *                                                                  *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS LIBRARY SOURCE IS     *
 * GOVERNED BY A BSD-STYLE SOURCE LICENSE INCLUDED WITH THIS SOURCE *
 * IN 'COPYING'. PLEASE READ THESE TERMS BEFORE DISTRIBUTING.       *
 *                                                                  *
 * THE OggVorbis 'TREMOR' SOURCE CODE IS (C) COPYRIGHT 1994-2002    *
 * BY THE Xiph.Org FOUNDATION http://www.xiph.org/                  *
 *                                                                  *
 ********************************************************************

 function: simple example decoder using vorbisidec

 ********************************************************************/

/* Takes a vorbis bitstream from stdin and writes raw stereo PCM to
   stdout using vorbisfile. Using vorbisfile is much simpler than
   dealing with libvorbis. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//#include <NDS.h>
#include "../../arm7/source/nds_compati.h"

#include "libmad/mad.h"

#include "filesys.h"

#include "boolean.h"

static void (*_consolePrintf)(const char* format, ...)=NULL;

void MP3_SetFunc_consolePrintf(u32 adr)
{
  _consolePrintf=(void*)adr;
}

// --------------------------------------------------------------------

#define meminfo_margin (1024)
#define meminfocount (0x400)
void* meminfo_ptr[meminfocount];
u32 meminfo_size[meminfocount];
u32 meminfo_TotalSize;
u32 meminfo_wlogflag;
u32 meminfo_waitflag;

void meminfo_Init(void)
{
  u32 cnt;
  
  for(cnt=0;cnt<meminfocount;cnt++){
    meminfo_ptr[cnt]=NULL;
    meminfo_size[cnt]=0;
  }
  
  meminfo_TotalSize=0;
}

void meminfo_Set(void *ptr,u32 size)
{
  if(ptr==NULL){
    _consolePrintf("meminfo_Set ptr==NULL\n");
    while(1);
  }
  if(size==0){
    _consolePrintf("meminfo_Set size==0\n");
    while(1);
  }
  
  u32 cnt;
  
  for(cnt=0;cnt<meminfocount;cnt++){
    if(meminfo_ptr[cnt]==ptr){
      _consolePrintf("meminfo_Set meminfo_ptr[cnt]==ptr\n");
      while(1);
    }
    if(meminfo_ptr[cnt]==NULL){
      meminfo_TotalSize+=size;
      meminfo_ptr[cnt]=ptr;
      meminfo_size[cnt]=size;
      return;
    }
  }
  
  _consolePrintf("meminfo_Set not freearea.\n");
  while(1);
}

void meminfo_Del(void *ptr)
{
  if(ptr==NULL){
    _consolePrintf("meminfo_Del ptr==NULL\n");
    while(1);
  }
  
  u32 cnt;
  
  for(cnt=0;cnt<meminfocount;cnt++){
    if(ptr==meminfo_ptr[cnt]){
      meminfo_TotalSize-=meminfo_size[cnt];
      meminfo_ptr[cnt]=NULL;
      meminfo_size[cnt]=0;
      return;
    }
  }
  
  _consolePrintf("meminfo_Del not fount.\n");
  while(1);
}

#define _KEYS           (*(volatile u16*)0x04000130)
#define _READ_KEYS      ((~(_KEYS)) & 0x3FF)

void meminfo_KeyWait(void)
{
  if(meminfo_waitflag!=0) return;
  
  while(_READ_KEYS!=0);
  while(_READ_KEYS==0);
}

#undef _KEYS
#undef _READ_KEYS

u32 meminfo_GetSize(void *ptr)
{
  if(ptr==NULL){
    _consolePrintf("meminfo_GetSize ptr==NULL\n");
    while(1);
  }
  
  u32 cnt;
  
  for(cnt=0;cnt<meminfocount;cnt++){
    if(ptr==meminfo_ptr[cnt]) return(meminfo_size[cnt]);
  }
  
  _consolePrintf("meminfo_GetSize not fount.\n");
  while(1);
}

void *safemalloc(size_t size)
{
  void *ptr;
  u32 adr;
  
  ptr=malloc(size);
  adr=(u32)ptr;
  
  if((adr&3)!=0){ // 4byteACgĂȂ
    free(ptr);
    _consolePrintf("malloc(%d) fail reallocerror.%x\n",size,(u32)ptr);
    while(1);
    return(NULL);
  }
  
  if(adr<0x02000000){ // 擪AhXCO
    free(ptr);
    _consolePrintf("malloc(%d) fail reallocerror.%x\n",size,(u32)ptr);
    while(1);
    return(NULL);
  }
  
  if((0x02000000+(4*1024*1024))<=(adr+size)){ // I[AhXCゾ
    free(ptr);
    _consolePrintf("malloc(%d) fail reallocerror.%x\n",size,(u32)ptr);
    while(1);
    return(NULL);
  }
  
  return(ptr);
}

static void *_safemalloc(size_t size)
{
  size+=meminfo_margin;
  
  if(meminfo_wlogflag==0) _consolePrintf("+%d",size);
  
  void *ptr;
  
  ptr=safemalloc(size);
  
  meminfo_Set(ptr,size);
  
  meminfo_KeyWait();
  return(ptr);
}

static void _safefree(void *ptr)
{
  if(meminfo_wlogflag==0) _consolePrintf("-");
  
  meminfo_Del(ptr);
  
  free(ptr);
  
  meminfo_KeyWait();
  return;
}

// --------------------------------------------------------------------

static FILE *FileHandle;
static s32 FileSize;
static s32 FileOffset;

typedef struct mad_decoder Tmad_decoder;

static Tmad_decoder StaticMadDecoder;

#define SamplePerFrame (1152)

// --------------------------------------------------------------------

static s16 *outlbuf,*outrbuf;
static u32 outofs;

#define INPUT_BUFFER_SIZE (32*1024)
static u8 *InputBuffer=NULL;

static
enum mad_flow input(void *data,
                    struct mad_stream *stream)
{
  size_t ReadSize,Remaining;
  u8 *ReadStart;
  
  if(stream->next_frame!=NULL)
  {
    Remaining=stream->bufend-stream->next_frame;
    memmove(InputBuffer,stream->next_frame,Remaining);
    ReadStart=InputBuffer+Remaining;
    ReadSize=INPUT_BUFFER_SIZE-Remaining;
    
    }else{
    ReadSize=INPUT_BUFFER_SIZE;
    ReadStart=InputBuffer;
    Remaining=0;
  }
  
  if((FileOffset+ReadSize)>=FileSize){
    ReadSize=FileSize-FileOffset;
  }
  
  if(ReadSize<=0) return MAD_FLOW_STOP;
  
  FileSys_fread(ReadStart,1,ReadSize,FileHandle);
  FileOffset+=ReadSize;
  
//  _consolePrintf("r%d->%d\n",Remaining,ReadSize);
  
  mad_stream_buffer(stream,InputBuffer,ReadSize+Remaining);
  
  stream->error=0;
  
  return MAD_FLOW_CONTINUE;
}

static inline
signed int scale(mad_fixed_t sample)
{
  /* round */
  sample += (1L << (MAD_F_FRACBITS - 16));

  /* clip */
  if (sample >= MAD_F_ONE)
    sample = MAD_F_ONE - 1;
  else if (sample < -MAD_F_ONE)
    sample = -MAD_F_ONE;

  /* quantize */
  return sample >> (MAD_F_FRACBITS + 1 - 16);
}

static
enum mad_flow output(void *data,
                     struct mad_header const *header,
                     struct mad_pcm *pcm)
{
  unsigned int nchannels, nsamples;
  mad_fixed_t const *left_ch, *right_ch;

  /* pcm->samplerate contains the sampling frequency */

  nchannels = pcm->channels;
  nsamples  = pcm->length;
  left_ch   = pcm->samples[0];
  right_ch  = pcm->samples[1];

  if(nsamples!=SamplePerFrame){
//    _consolePrintf("nsamples==%d!=%d\n",nsamples,SamplePerFrame);
//    return MAD_FLOW_STOP;
  }
  
  if((outlbuf==NULL)||(outrbuf==NULL)){
    outofs+=nsamples;
    return MAD_FLOW_CONTINUE;
  }
  
  if (nchannels == 1) {
    while (nsamples--) {
      signed int sample;

      sample = scale(*left_ch++);
      outlbuf[outofs]=sample;
      outrbuf[outofs]=sample;
      outofs++;
    }
    return MAD_FLOW_CONTINUE;
  }

  if (nchannels == 2) {
    while (nsamples--) {
      signed int sample;

      sample = scale(*left_ch++);
      outlbuf[outofs]=sample;
      sample = scale(*right_ch++);
      outrbuf[outofs]=sample;
      outofs++;
    }
    return MAD_FLOW_CONTINUE;
  }

  return MAD_FLOW_STOP;
}

static
enum mad_flow error(void *data,
                    struct mad_stream *stream,
                    struct mad_frame *frame)
{
  _consolePrintf("decoding error 0x%04x (%s) at byte offset %u\n",
          stream->error, mad_stream_errorstr(stream),
          FileOffset);

  /* return MAD_FLOW_BREAK here to stop decoding (and propagate an error) */

  return MAD_FLOW_CONTINUE;//MAD_FLOW_STOP;
}

// --------------------------------------------------------------------

#define madusr_Return_OK (0)
#define madusr_Return_NG (-1)

#define madusr_ExecReturn_Next (0)
#define madusr_ExecReturn_End (-1)
#define madusr_ExecReturn_Fail (-2)

static
int madusr_decode_init(struct mad_decoder *decoder)
{
  decoder->sync = NULL;
  
  /* configure input, output, and error functions */
  
  mad_decoder_init(decoder, NULL, input, 0 /* header */, 0 /* filter */, output, error, 0 /* message */);
  
  if (decoder->input_func == 0) return madusr_Return_NG;
  if (decoder->output_func == 0) return madusr_Return_NG;
  if (decoder->error_func==0) return madusr_Return_NG;

  struct mad_stream *stream;
  struct mad_frame *frame;
  struct mad_synth *synth;
  
  decoder->sync = _safemalloc(sizeof(*decoder->sync));
  if (decoder->sync == NULL) return madusr_Return_NG;
  
  stream = &decoder->sync->stream;
  frame  = &decoder->sync->frame;
  synth  = &decoder->sync->synth;

  mad_stream_init(stream);
  mad_frame_init(frame);
  mad_synth_init(synth);

  mad_stream_options(stream, decoder->options);
  
  if(decoder->input_func(decoder->cb_data, stream)!=MAD_FLOW_CONTINUE) return madusr_Return_NG;
  
  return madusr_Return_OK;
}

static
int madusr_decode_exec(struct mad_decoder *decoder)
{
  enum mad_flow (*error_func)(void *, struct mad_stream *, struct mad_frame *);
  void *error_data;
  
  struct mad_stream *stream;
  struct mad_frame *frame;
  struct mad_synth *synth;

  error_func = decoder->error_func;
  error_data = decoder->cb_data;
  
  stream = &decoder->sync->stream;
  frame  = &decoder->sync->frame;
  synth  = &decoder->sync->synth;

  if(stream->error == MAD_ERROR_BUFLEN){
    return madusr_ExecReturn_Fail;
  }
  
  while(1){
    if (decoder->header_func) {
      if (mad_header_decode(&frame->header, stream) == -1) {
        if (!MAD_RECOVERABLE(stream->error)){
          if(stream->error==MAD_ERROR_BUFLEN){
            if(decoder->input_func(decoder->cb_data, stream)!=MAD_FLOW_CONTINUE) return madusr_ExecReturn_Fail;
            continue;
          }
          return madusr_ExecReturn_Fail;
        }
  
        if(stream->error==MAD_ERROR_BUFLEN){
          if(decoder->input_func(decoder->cb_data, stream)!=MAD_FLOW_CONTINUE) return madusr_ExecReturn_Fail;
          continue;
        }
        
        switch (error_func(error_data, stream, frame)) {
          case MAD_FLOW_STOP:
            return madusr_ExecReturn_End;
          case MAD_FLOW_BREAK:
            return madusr_ExecReturn_Fail;
          case MAD_FLOW_IGNORE:
          case MAD_FLOW_CONTINUE:
          default:
            return madusr_ExecReturn_Next;
        }
      }
  
      switch (decoder->header_func(decoder->cb_data, &frame->header)) {
        case MAD_FLOW_STOP:
          return madusr_ExecReturn_End;
        case MAD_FLOW_BREAK:
          return madusr_ExecReturn_Fail;
        case MAD_FLOW_IGNORE:
          return madusr_ExecReturn_Next;
        case MAD_FLOW_CONTINUE:
          continue;
          break;
      }
    }
    break;
  }
  
  while(1){
    if (mad_frame_decode(frame, stream) == -1) {
    
      if (!MAD_RECOVERABLE(stream->error)){
        if(stream->error==MAD_ERROR_BUFLEN){
          if(decoder->input_func(decoder->cb_data, stream)!=MAD_FLOW_CONTINUE) return madusr_ExecReturn_Fail;
          continue;
        }
        return madusr_ExecReturn_Fail;
      }
      
      switch (error_func(error_data, stream, frame)) {
        case MAD_FLOW_STOP:
          return madusr_ExecReturn_End;
        case MAD_FLOW_BREAK:
          return madusr_ExecReturn_Fail;
        case MAD_FLOW_IGNORE:
          return madusr_ExecReturn_Next;
        case MAD_FLOW_CONTINUE:
          continue;
        default:
          break;
      }
    }
    break;
  }
  
  if((outlbuf==NULL)||(outrbuf==NULL)){
    outofs=1;
    return madusr_ExecReturn_Next;
  }
  
  if (decoder->filter_func) {
    switch (decoder->filter_func(decoder->cb_data, stream, frame)) {
      case MAD_FLOW_STOP:
        return madusr_ExecReturn_End;
      case MAD_FLOW_BREAK:
        return madusr_ExecReturn_Fail;
      case MAD_FLOW_IGNORE:
        return madusr_ExecReturn_Next;
      case MAD_FLOW_CONTINUE:
        break;
    }
  }

  mad_synth_frame(synth, frame);

  if (decoder->output_func) {
    switch (decoder->output_func(decoder->cb_data, &frame->header, &synth->pcm)) {
      case MAD_FLOW_STOP:
        return madusr_ExecReturn_End;
      case MAD_FLOW_BREAK:
        return madusr_ExecReturn_Fail;
      case MAD_FLOW_IGNORE:
      case MAD_FLOW_CONTINUE:
        break;
    }
  }
  
  return madusr_ExecReturn_Next;
}

static
void madusr_decode_free(struct mad_decoder *decoder)
{
  struct mad_stream *stream;
  struct mad_frame *frame;
  struct mad_synth *synth;
  
  if(decoder->sync!=NULL){
    stream = &decoder->sync->stream;
    frame  = &decoder->sync->frame;
    synth  = &decoder->sync->synth;
    
    mad_synth_finish(synth);
    mad_frame_finish(frame);
    mad_stream_finish(stream);
    
    _safefree(decoder->sync);
    decoder->sync = NULL;
  }
  
  mad_decoder_finish(decoder);
}

// --------------------------------------------------------------------

// --------------------------------------------------------------------

static b8 Initialized=False;

void FreeMP3(void);

b8 StartMP3(FILE *_FileHandle)
{
  meminfo_Init();
  
  meminfo_wlogflag=1;
  meminfo_waitflag=1;
  
  if(Initialized==True){
    FreeMP3();
  }
  Initialized=True;
  
  FileHandle=_FileHandle;
  FileOffset=0;
  
  FileSys_fseek(FileHandle,0,SEEK_END);
  FileSize=FileSys_ftell(FileHandle);
  FileSys_fseek(FileHandle,0,SEEK_SET);

  InputBuffer=(u8*)_safemalloc(INPUT_BUFFER_SIZE+MAD_BUFFER_GUARD);
  if(InputBuffer==NULL){
    _consolePrintf("InputBuffer out of memory.\n");
    FreeMP3();
    return(False);
  }
  
  _consolePrintf("libmad init.\n");
  
  int result;
  
  result=madusr_decode_init(&StaticMadDecoder);
  if(result==madusr_Return_NG){
    FreeMP3();
    return(False);
  }
  
  outlbuf=NULL;
  outrbuf=NULL;
  outofs=0;
  
  if(madusr_decode_exec(&StaticMadDecoder)!=madusr_ExecReturn_Next){
    FreeMP3();
    return(False);
  }
  
  struct mad_frame *frame;
  
  frame  = &StaticMadDecoder.sync->frame;
  
  _consolePrintf("\n");
  
  _consolePrintf("Format=");
  switch(frame->header.layer){
    case MAD_LAYER_I:
      _consolePrintf("Layer I\n");
      break;
    case MAD_LAYER_II:
      _consolePrintf("Layer II\n");
      break;
    case MAD_LAYER_III:
      _consolePrintf("Layer III\n");
      break;
    default:
      _consolePrintf("unknown layer.\n");
      FreeMP3();
      return(False);
      break;
  }

  _consolePrintf("ChannelMode=");
  switch(frame->header.mode){
    case MAD_MODE_SINGLE_CHANNEL:
      _consolePrintf("SingleChannel\n");
      break;
    case MAD_MODE_DUAL_CHANNEL:
      _consolePrintf("DualChannel\n");
      break;
    case MAD_MODE_JOINT_STEREO:
      _consolePrintf("JointStereo\n");
      break;
    case MAD_MODE_STEREO:
      _consolePrintf("Normal LR Stereo\n");
      break;
    default:
      _consolePrintf("unknown ChannelMode.\n");
      FreeMP3();
      return(False);
      break;
  }
  
  _consolePrintf("BitRate=%dkbps\n",frame->header.bitrate/1000);
  _consolePrintf("SampleRate=%dHz\n",frame->header.samplerate);
  
  return(True);
}

u32 UpdateMP3(s16 *lbuf,s16 *rbuf)
{
  if(Initialized==False) return(0);
  
  int result;
  
  outlbuf=lbuf;
  outrbuf=rbuf;
  outofs=0;
  
  result=madusr_ExecReturn_Next;
  
  while((result==madusr_ExecReturn_Next)&&(outofs<SamplePerFrame)){
    result=madusr_decode_exec(&StaticMadDecoder);
  }
  
  if(result==madusr_ExecReturn_Next){
    return(outofs);
  }
  
  if(result==madusr_ExecReturn_End){
//    _consolePrintf("ExecReturn_End\n");
  }
  if(result==madusr_ExecReturn_Fail){
//    _consolePrintf("ExecReturn_Fail\n");
  }
  
  return(0);
}

void FreeMP3(void)
{
  if(Initialized==True){
    madusr_decode_free(&StaticMadDecoder);
    if(InputBuffer!=NULL){
      _safefree(InputBuffer); InputBuffer=NULL;
    }
  }
  Initialized=False;
}

s32 MP3_GetFileSize(void)
{
  return(FileSize);
}

s32 MP3_GetFileOffset(void)
{
  return(FileOffset);
}

void MP3_SetFileOffset(s32 ofs)
{
  if(FileSize<=ofs) ofs=FileSize-1;
  if(ofs<0) ofs=0;
  
  FileOffset=ofs;
  if(FileHandle!=NULL) FileSys_fseek(FileHandle,FileOffset,SEEK_SET);
}

u32 MP3_GetBitRate(void)
{
  struct mad_frame *frame;
  
  frame  = &StaticMadDecoder.sync->frame;
  
  return(frame->header.bitrate);
}

u32 MP3_GetChannelCount(void)
{
  struct mad_frame *frame;
  
  frame  = &StaticMadDecoder.sync->frame;
  
  switch(frame->header.mode){
    case MAD_MODE_SINGLE_CHANNEL:
      return(1);
      break;
    case MAD_MODE_DUAL_CHANNEL:
      return(2);
      break;
    case MAD_MODE_JOINT_STEREO:
      return(2);
      break;
    case MAD_MODE_STEREO:
      return(2);
      break;
    default:
      return(1);
      break;
  }
}

u32 MP3_GetSampleRate(void)
{
  struct mad_frame *frame;
  
  frame  = &StaticMadDecoder.sync->frame;
  
  return(frame->header.samplerate);
}

u32 MP3_GetSamplePerFrame(void)
{
  return(SamplePerFrame);
}

