
#include <nds.h>
#include "../../arm7/source/nds_compati.h"
#include <NDS/ARM9/CP15.h>


#include <stdio.h>
#include <stdlib.h>

#include "memtool.h"

#include "../../ipc3.h"

#include "strpcm.h"

#include "filesys.h"

static volatile b8 strpcmPause;

volatile b8 VsyncPassed=False;

volatile b8 strpcmRequestStop;

volatile u32 strpcmRingBufReadIndex;
volatile u32 strpcmRingBufWriteIndex;

s16 *strpcmRingLBuf=NULL;
s16 *strpcmRingRBuf=NULL;


u32 strpcmVolume8;

void InterruptHandler(void)
{
  u32 flags = IF;

  if(flags&IRQ_IPC_SYNC){
    strpcmUpdate();
    
    u32 Samples=IPC3->strpcmSamples;
    
    s16 *ldst=IPC3->strpcmLBuf;
    s16 *rdst=IPC3->strpcmRBuf;
    
    DC_FlushRangeOverrun(ldst,Samples*2);
    DC_FlushRangeOverrun(rdst,Samples*2);
    
    IF = IRQ_IPC_SYNC;
  }

  if (flags & IRQ_VBLANK) {
  	extern vu32 vs_count;
	vs_count++;

    VsyncPassed=True;
    
    VBLANK_INTR_WAIT_FLAGS|=IRQ_VBLANK;
    IF = IRQ_VBLANK;
  }
}


void strpcmStart(u32 SampleRate,u32 SamplePerBuf,u32 ChannelCount)
{
  DC_FlushAll();
  while(IPC3->strpcmControl!=strpcmControl_NOP){
    DC_FlushAll();
  }
  
  strpcmRequestStop=False;
  strpcmPause=False;
  
  u32 Samples=SamplePerBuf;
  u32 RingSamples=Samples*strpcmRingBufCount;
  
  strpcmRingBufReadIndex=0;
  strpcmRingBufWriteIndex=0;
  
  strpcmRingLBuf=(s16*)safemalloc(RingSamples*2);
  strpcmRingRBuf=(s16*)safemalloc(RingSamples*2);
  
  {
    u32 cnt;
    for(cnt=0;cnt<RingSamples;cnt++){
      strpcmRingLBuf[cnt]=0;
      strpcmRingRBuf[cnt]=0;
    }
  }
  
  IPC3->strpcmFreq=SampleRate;
  IPC3->strpcmSamples=Samples;
  IPC3->strpcmChannels=ChannelCount;
  
  IPC3->strpcmLBuf=(s16*)safemalloc(Samples*2);
  IPC3->strpcmRBuf=(s16*)safemalloc(Samples*2);
  
  {
    u32 cnt;
    for(cnt=0;cnt<Samples;cnt++){
      IPC3->strpcmLBuf[cnt]=0;
      IPC3->strpcmRBuf[cnt]=0;
    }
  }
  
  DC_FlushAll();
  IPC3->strpcmControl=strpcmControl_Play;
  
  DC_FlushAll();
  while(IPC3->strpcmControl!=strpcmControl_NOP){
    DC_FlushAll();
  }
}

#define DISP_SR        (*(vuint16*)0x04000004)
void InitInterrupts(void)
{
	extern	u32 vs_bios_flg;

  IME = 0;
  IRQ_HANDLER = &InterruptHandler;

  IE = IRQ_IPC_SYNC | IRQ_VBLANK;

  DISP_SR = DISP_VBLANK_IRQ;

  IPC_SYNC=IPC_SYNC_IRQ_ENABLE;
  
  IF = ~0;
  IME = 1;
  
  vs_bios_flg=1;

}

void strpcmStop(void)
{
  strpcmRequestStop=True;
  
  DC_FlushAll();
  while(IPC3->strpcmControl!=strpcmControl_NOP){
    DC_FlushAll();
  }
  
  IPC3->strpcmControl=strpcmControl_Stop;
  
  DC_FlushAll();
  while(IPC3->strpcmControl!=strpcmControl_NOP){
    DC_FlushAll();
  }
  
  strpcmRequestStop=False;
  strpcmPause=False;
  
  strpcmRingBufReadIndex=0;
  strpcmRingBufWriteIndex=0;
  
  if(strpcmRingLBuf!=NULL){
    safefree((void*)strpcmRingLBuf); strpcmRingLBuf=NULL;
  }
  if(strpcmRingRBuf!=NULL){
    safefree((void*)strpcmRingRBuf); strpcmRingRBuf=NULL;
  }
  
  IPC3->strpcmFreq=0;
  IPC3->strpcmSamples=0;
  IPC3->strpcmChannels=0;
  
  if(IPC3->strpcmLBuf!=NULL){
    safefree(IPC3->strpcmLBuf); IPC3->strpcmLBuf=NULL;
  }
  if(IPC3->strpcmRBuf!=NULL){
    safefree(IPC3->strpcmRBuf); IPC3->strpcmRBuf=NULL;
  }
}
u8 ocnt;
void strpcmUpdate(void)
{
  b8 IgnoreFlag=False;
  
  u32 Samples=IPC3->strpcmSamples;
  
  s16 *ldst=IPC3->strpcmLBuf;
  s16 *rdst=IPC3->strpcmRBuf;
  
  u32 CurIndex=(strpcmRingBufReadIndex+1) & strpcmRingBufBitMask;
  
  s16 *lsrc=&strpcmRingLBuf[Samples*CurIndex];
  s16 *rsrc=&strpcmRingRBuf[Samples*CurIndex];
  
  if(strpcmPause==True) IgnoreFlag=True;
  if((ldst==NULL)||(rdst==NULL)) IgnoreFlag=True;
  if(strpcmRequestStop==True) IgnoreFlag=True;
  
  if(CurIndex==strpcmRingBufWriteIndex) IgnoreFlag=True;
  if((lsrc==NULL)||(rsrc==NULL)) IgnoreFlag=True;
  
  if(IgnoreFlag==True){
    MemSet16DMA3(0,ldst,Samples*2);
    MemSet16DMA3(0,rdst,Samples*2);
    return;
  }
 	
 	u32 cnt = 0;
  for( cnt=0;cnt<Samples;cnt++){
    ldst[cnt]=lsrc[cnt];
    rdst[cnt]=rsrc[cnt];
   }
  
//  MemCopy16DMA3(lsrc,ldst,Samples*2);
//  MemCopy16DMA3(rsrc,rdst,Samples*2);
  
  strpcmRingBufReadIndex=CurIndex;
}

void strpcmSetVolume8(u32 v)
{
  strpcmVolume8=v;
  
  IPC3->strpcmVolume8=strpcmVolume8;
}

u32 strpcmGetVolume8(void)
{
  return(strpcmVolume8);
}

void strpcmSetPause(b8 v)
{
  strpcmPause=v;
}

b8 strpcmGetPause(void)
{
  return(strpcmPause);
}

