//////////////////////////////////////////////////////////////////////
//
// ipc.h -- Inter-processor communication
//
// version 0.1, February 14, 2005
//
//  Copyright (C) 2005 Michael Noland (joat) and Jason Rogers (dovoto)
//
//  This software is provided 'as-is', without any express or implied
//  warranty.  In no event will the authors be held liable for any
//  damages arising from the use of this software.
//
//  Permission is granted to anyone to use this software for any
//  purpose, including commercial applications, and to alter it and
//  redistribute it freely, subject to the following restrictions:
//
//  1. The origin of this software must not be misrepresented; you
//     must not claim that you wrote the original software. If you use
//     this software in a product, an acknowledgment in the product
//     documentation would be appreciated but is not required.
//  2. Altered source versions must be plainly marked as such, and
//     must not be misrepresented as being the original software.
//  3. This notice may not be removed or altered from any source
//     distribution.
//
// Changelog:
//   0.1: First version
//
//////////////////////////////////////////////////////////////////////

#ifndef NDS_IPC3_INCLUDE
#define NDS_IPC3_INCLUDE

//////////////////////////////////////////////////////////////////////

#include <nds.h>
#include <nds/jtypes.h>
#include <nds/ipc.h>

//////////////////////////////////////////////////////////////////////

typedef struct sTransferRegion3 {
  uint32 heartbeat;          // counts frames
 
   int16 touchX,   touchY;   // TSC X, Y
   int16 touchXpx, touchYpx; // TSC X, Y pixel values
   int16 touchZ1,  touchZ2;  // TSC x-panel measurements
  uint16 tdiode1,  tdiode2;  // TSC temperature diodes
  uint32 temperature;        // TSC computed temperature
  
 
  uint16 buttons;            // X, Y, /PENIRQ buttons
 
  union {
		vuint8 curtime[8];		// current time response from RTC

		struct {
			vu8 command;
			vu8 year;		//add 2000 to get 4 digit year
			vu8 month;		//1 to 12
			vu8 day;			//1 to (days in month)

			vu8 weekday;			// day of week
			vu8 hours;		//0 to 11 for AM, 52 to 63 for PM
			vu8 minutes;		//0 to 59
			vu8 seconds;		//0 to 59
		} rtc;
	} time;
 
  uint16 battery;            // battery life ??  hopefully.  :)
  uint16 aux;                // i have no idea...
  
  // Don't rely on these below, will change or be removed in the future
  vuint32 mailAddr;
  vuint32 mailData;
  vuint8 mailRead;
  vuint8 mailBusy;
  vuint8 mailSize;
  
  u32 ReqVsyncUpdate;
  
  u32 strpcmControl;
  u32 strpcmFreq,strpcmSamples,strpcmChannels;
  u32 strpcmVolume8;
  s16 *strpcmLBuf,*strpcmRBuf;
  
  
} TransferRegion3, * pTransferRegion3;

//////////////////////////////////////////////////////////////////////

#define IPC3 ((TransferRegion3 volatile *)(0x027FF000))

#define strpcmControl_NOP (0)
#define strpcmControl_Play (1)
#define strpcmControl_Stop (2)

#endif

//////////////////////////////////////////////////////////////////////

